
/*********************************************************************
 Copyright (C) 2001-2009
 Sigma Designs, Inc. 

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License version 2 as
 published by the Free Software Foundation.
 *********************************************************************/

/*
 * arch/mips/tangox/delay.c
 *
 * Copyright (C) 2003-2009 Sigma Designs, Inc
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>
#include <linux/irq.h>
#include <linux/sched.h>

#include "setup.h"

/*
 * helpers to access cpu block registers
 */
#define RD_CPU_REG32(r)	\
		gbus_read_reg32(REG_BASE_cpu_block + (r))

#define WR_CPU_REG32(r, v)	\
		gbus_write_reg32(REG_BASE_cpu_block + (r), (v))

typedef int (*COND_FUNC_PTR)(void *);

static unsigned long timer_status = 0;

//#define TANGOX_USE_XTAL_UDELAY

static inline unsigned long tangox_get_timestamp(void)
{
#ifdef TANGOX_USE_XTAL_UDELAY
	return(gbus_read_reg32(REG_BASE_system_block + SYS_xtal_in_cnt));
#else
	return(read_c0_count());
#endif
}
static inline unsigned long tangox_get_timecount(unsigned usec)
{
#ifdef TANGOX_USE_XTAL_UDELAY
	return(usec * (TANGOX_BASE_FREQUENCY / 1000000));
#else
	extern unsigned long em8xxx_cpu_frequency;
	unsigned long cnt = em8xxx_cpu_frequency / (1000000 * 2);
	if ((cnt * (1000000 * 2)) < em8xxx_cpu_frequency)
		cnt++; /* take the ceiling */
	return(usec * cnt);
#endif
}

static inline int tangox_time_in_range(unsigned long start, unsigned long end, unsigned long timestamp)
{
	if (end > start)
		return((timestamp < end) && (timestamp >= start)); 
	else
		return((timestamp < end) || (timestamp >= start));
}

static inline int tangox_wait_until(unsigned long start, unsigned long end, COND_FUNC_PTR func_ptr, void *arg)
{
	unsigned long timestamp;

	if (func_ptr) {
		for (timestamp = tangox_get_timestamp(); tangox_time_in_range(start, end, timestamp); timestamp = tangox_get_timestamp()) {
			if (!((*func_ptr)(arg)))
				return 0; /* wait until it's out of range, or condition evaluation is false */
		}
	} else {
		for (timestamp = tangox_get_timestamp(); tangox_time_in_range(start, end, timestamp); timestamp = tangox_get_timestamp()) {
			; /* wait until it's out of range */
		}
	}

	return 1; /* time's up and condition is still true */
}

static inline int tangox_udelay_poll_with_condition(unsigned usec, COND_FUNC_PTR func_ptr, void *arg)
{
	unsigned long start = tangox_get_timestamp();
	unsigned long end = start + tangox_get_timecount(usec);
	return tangox_wait_until(start, end, func_ptr, arg);
}

static inline int tangox_check_time_range_with_condition(unsigned long start, unsigned long end, COND_FUNC_PTR func_ptr, void *arg) 
{
	if (func_ptr)
		return(tangox_time_in_range(start, end, tangox_get_timestamp()) && (*func_ptr)(arg));
	else
		return(tangox_time_in_range(start, end, tangox_get_timestamp()));
}

static inline int tangox_check_time_range(unsigned long start, unsigned long end)
{
	return(tangox_check_time_range_with_condition(start, end, NULL, NULL));
}

#ifndef CONFIG_TANGOX_USE_CPU_CLOCK

#define MAX_POLL_TIME		(25)	/* less than this we use busy polling */
#define max_val(x,y)		(((x)>=(y))?(x):(y))
static DECLARE_WAIT_QUEUE_HEAD(cp0_timer_wq);
static int cp0_timer_expired = 0, cp0_timer_init = 0, cp0_timer_active = 0;
static spinlock_t cp0_timer_lock;		/* Spin lock */
static int mips_timer_irq = 0;
static unsigned int cp0_start, cp0_end;

static irqreturn_t cp0_timer_isr(int irq, void *dev_id)
{
	wait_queue_head_t *q = (wait_queue_head_t *)dev_id;
	unsigned int c0_count = read_c0_count();
	unsigned long flags;
	if (q != &cp0_timer_wq)		/* Paranoid check */
		return IRQ_NONE;

	spin_lock_irqsave(&cp0_timer_lock, flags);
	if (cp0_timer_active) {
		if (tangox_time_in_range(cp0_start, cp0_end, c0_count)) {
			write_c0_compare(cp0_end);
		} else {
			write_c0_compare(read_c0_count() - 1);
			cp0_timer_active = 0;
			cp0_timer_expired = 1;
			wake_up_interruptible(&cp0_timer_wq);
		}
	} else 
		write_c0_compare(read_c0_count() - 1);

	spin_unlock_irqrestore(&cp0_timer_lock, flags);
	return IRQ_HANDLED;
}

int tangox_udelay_with_condition(unsigned int usecond, COND_FUNC_PTR func_ptr, void *arg)
{
	unsigned long start = tangox_get_timestamp();
	unsigned long end = start + tangox_get_timecount(usecond);
	unsigned long flags;
	unsigned int wait_count, usec;

	if (usecond == 0)
		return 1;
	else if (in_atomic()) 
		goto poll;	/* not safe for context switching */
	else if ((usecond < MAX_POLL_TIME) || (func_ptr == NULL))
		goto poll;
	else if (unlikely((cp0_timer_init == 0) || (cp0_timer_active != 0))) 
		goto poll;
	else if (test_and_set_bit(0, &timer_status) != 0)
		goto poll;

	while (usecond >= MAX_POLL_TIME) { 
		usec = max_val(usecond - MAX_POLL_TIME, MAX_POLL_TIME);
		wait_count = tangox_get_timecount(usec);
		spin_lock_irqsave(&cp0_timer_lock, flags);
		cp0_timer_active = 1;
		cp0_start = read_c0_count();
		cp0_end = cp0_start + wait_count;
		write_c0_compare(cp0_end);
		spin_unlock_irqrestore(&cp0_timer_lock, flags);

		wait_event_interruptible(cp0_timer_wq, cp0_timer_expired != 0);
		cp0_timer_expired = 0;

		if (func_ptr) {
			if (!((*func_ptr)(arg)))
				goto done;
		}

		usecond -= usec;
	}
	clear_bit(0, &timer_status);

poll:
	return tangox_wait_until(start, end, func_ptr, arg);

done:
	clear_bit(0, &timer_status);
	return 0;
}

int __init tangox_cp0_timer_init(void)
{
	spin_lock_init(&cp0_timer_lock);
#ifdef CONFIG_TANGO2
	mips_timer_irq = 7;
#elif defined(CONFIG_TANGO3)
	mips_timer_irq = (read_c0_intctl() >> 29) & 7;
#endif

	write_c0_compare(read_c0_count() - 1);
	if (request_irq(mips_timer_irq, cp0_timer_isr, IRQF_DISABLED, "cp0_timer", 
			&cp0_timer_wq) != 0) {
		printk(KERN_ERR "cp0_timer: cannot register IRQ (%d)\n", mips_timer_irq);
		return -EIO;
	} else {
		cp0_timer_init = 1;
		printk(KERN_INFO "cp0_timer: interrupt registered.\n");
	}
	return 0;
}

__initcall(tangox_cp0_timer_init);

#else
/*
 * CPU_time0_load
 *  clock / HZ / (2 * prescale)
 * CPU_time0_ctrl
 *  PS(D2-3) : prescale. 0x00 = 1, 0x01 = 16, 0x10 = 256
 *    There is a bug, and the actual prescale is 0x01 = 32, 0x10 = 512
 *  M(D6) : periodic mode
 *  E(D7) : enable
 */
#define TIMER_ENABLE            0x80    // D7
#define TIMER_PERIODIC          0x40    // D6
#define TIMER_PRESCALE_1        0x00    // D[2-3] = 00b
#define TIMER_PRESCALE_32       0x04    // D[2-3] = 01b
#define TIMER_PRESCALE_512      0x08    // D[2-3] = 10b

#define MAX_POLL_TIME		(25)	/* less than this we use busy polling */
#define max_val(x,y)		(((x)>=(y))?(x):(y))

static const unsigned timer_prescale_bit[3] = { 0, 5, 9 };
static const unsigned timer_prescale[3] = { TIMER_PRESCALE_1, TIMER_PRESCALE_32, TIMER_PRESCALE_512 };

static DECLARE_WAIT_QUEUE_HEAD(timer0_wq);
static int timer0_expired = 0, timer0_init = 0;
static spinlock_t timer0_lock;		/* Spin lock */

static irqreturn_t timer0_isr(int irq, void *dev_id)
{
	wait_queue_head_t *q = (wait_queue_head_t *)dev_id;
	unsigned long flags;
	if (q != &timer0_wq)		/* Paranoid check */
		return IRQ_NONE;

	spin_lock_irqsave(&timer0_lock, flags);
	WR_CPU_REG32(CPU_time0_clr, 1);
	WR_CPU_REG32(CPU_time0_ctrl, 0); /* disable timer */
	timer0_expired = 1;
	spin_unlock_irqrestore(&timer0_lock, flags);

	wake_up_interruptible(&timer0_wq);

	return IRQ_HANDLED;
}

/*
 * udelay with timer0 implementation if possible
 */
int tangox_udelay_with_condition(unsigned int usecond, COND_FUNC_PTR func_ptr, void *arg)
{
	unsigned long start = tangox_get_timestamp();
	unsigned long end = start + tangox_get_timecount(usecond);
	unsigned long loop_end;
	unsigned long usec;
	unsigned int timer_val, tmp;
	unsigned long flags;
	int idx;

	if (usecond == 0)
		return 1;
	else if (in_atomic()) 
		goto poll;	/* not safe for context switching */
	else if ((usecond < MAX_POLL_TIME) || (func_ptr == NULL))
		goto poll;
	else if (unlikely(timer0_init == 0)) 
		goto poll;
	else if (test_and_set_bit(0, &timer_status) != 0)
		goto poll;

	loop_end = start + ((usecond >= MAX_POLL_TIME) ? tangox_get_timecount(usecond - MAX_POLL_TIME) : 0);
	while ((usecond >= MAX_POLL_TIME) && tangox_time_in_range(start, loop_end, tangox_get_timestamp())) { 
		usec = max_val(usecond - MAX_POLL_TIME, MAX_POLL_TIME);
		spin_lock_irqsave(&timer0_lock, flags);
		if (RD_CPU_REG32(CPU_time0_ctrl) & TIMER_ENABLE) {
			spin_unlock_irqrestore(&timer0_lock, flags);
			if (!tangox_udelay_poll_with_condition(usec, func_ptr, arg)); /* timer0 is in used */
				goto done;
			goto next;
		}
		/* CPU_time0_value register contains just 16-bits value So
		   take care not to let the value to overflow */
		for (idx = 0; idx < 3; idx++) {
			tmp = (usec * (em8xxx_sys_frequency / 1000)) / 1000; /* usec needs to be <= 10000 */
			timer_val = tmp >> timer_prescale_bit[idx];
			if ((timer_val & 0xffff0000) == 0) {
				if (tmp & ((1 << timer_prescale_bit[idx]) - 1))
					timer_val++;
				break;
			}
		}
		if (idx >= 3) {
			spin_unlock_irqrestore(&timer0_lock, flags);
			if (!tangox_udelay_poll_with_condition(usec, func_ptr, arg))
				goto done;
			goto next;
		}

		WR_CPU_REG32(CPU_time0_load, timer_val);
		WR_CPU_REG32(CPU_time0_ctrl, TIMER_ENABLE | TIMER_PERIODIC | timer_prescale[idx]);
		spin_unlock_irqrestore(&timer0_lock, flags);

		/* sleep in the wait queue until timer expired, or timeout */
		wait_event_interruptible(timer0_wq, timer0_expired != 0);
		timer0_expired = 0;

		WR_CPU_REG32(CPU_time0_load, 0);

		if (func_ptr) {
			if (!((*func_ptr)(arg)))
				goto done;
		}
next:
		usecond -= usec;
		continue;
	}
	clear_bit(0, &timer_status);

poll:
	return tangox_wait_until(start, end, func_ptr, arg);

done:
	clear_bit(0, &timer_status);
	return 0;
}

int __init tangox_timer0_init(void)
{
	WR_CPU_REG32(CPU_time0_ctrl, 0); /* disable timer to begin with */
	WR_CPU_REG32(CPU_time0_load, 0);
	WR_CPU_REG32(CPU_time0_clr, 1);
	spin_lock_init(&timer0_lock);

	if (request_irq(LOG2_CPU_TIMER0_INT + IRQ_CONTROLLER_IRQ_BASE, timer0_isr, IRQF_DISABLED, "timer0", 
			&timer0_wq) != 0) {
		printk(KERN_ERR "timer0: cannot register IRQ (%d)\n", LOG2_CPU_TIMER0_INT + IRQ_CONTROLLER_IRQ_BASE);
		return -EIO;
	} else {
		timer0_init = 1;
		printk(KERN_INFO "timer0: interrupt registered.\n");
	}
	return 0;
}

__initcall(tangox_timer0_init);

#endif /* !CONFIG_TANGOX_USE_CPU_CLOCK */

void tangox_udelay(unsigned int usec)
{
	tangox_udelay_with_condition(usec, NULL, NULL);
}

EXPORT_SYMBOL(tangox_udelay);
EXPORT_SYMBOL(tangox_udelay_with_condition);

