/* linux/drivers/mtd/nand/smp86xx.c
 *
 * Copyright (c) 2009 Sigma Designs
 *
 * SMP86xx NAND driver
 *
 * Changelog:
 *	23-Mar-2010  YH	  Work with NAND 1.2.1 or above
 *	09-Jun-2009	  Add Xenv config
 *	12-May-2009  BJD  Initial version
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/ioport.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <linux/slab.h>
#include <linux/clk.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/ctype.h>

#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ecc.h>
#include <linux/mtd/partitions.h>

#include <asm/io.h>

#include <asm/tango3/rmdefs.h>
#include <asm/tango3/tango3.h>
#include <asm/tango3/tango3api.h>

/**
	NAND access register
*/
#define SMP86XX_REG_CMD				0
#define SMP86XX_REG_ADDR			4
#define SMP86XX_REG_DATA			8

#define tReset					5 //us

#define PB_IORDY	0x80000000

/**
	PB Chip Select 
*/
#define	MAX_NO_CS			8

/**
	gbus access micro
*/
#define RD_HOST_REG32(r)	\
		gbus_read_reg32(REG_BASE_host_interface + (r))

#define WR_HOST_REG32(r, v)	\
		gbus_write_reg32(REG_BASE_host_interface + (r), (v))

#define RD_HOST_REG16(r)	\
		gbus_read_reg16(REG_BASE_host_interface + (r))

#define WR_HOST_REG16(r, v) \
		gbus_write_reg16(REG_BASE_host_interface + (r), (v))

#define RD_HOST_REG8(r)	\
		gbus_read_reg8(REG_BASE_host_interface + (r))

#define WR_HOST_REG8(r, v)	\
		gbus_write_reg8(REG_BASE_host_interface + (r), (v))

#define SMP_NAND_DEV_NAME	"[SMP_NAND]"
#define SMP_NAND_DRV_VERSION	"0.2"

#define MAX_CS		8	/* Maximum number of CS */
#define MAX_PARTITIONS	16	/* Maximum partitions per CS */

/* XENV keys to be used */
#define CS_RESERVED	"a.cs%d_rsvd_pblk"
#define CS_PARTS	"a.cs%d_pblk_parts"
#define CS_PART_SIZE	"a.cs%d_pblk_part%d_size"
#define CS_PART_OFFSET	"a.cs%d_pblk_part%d_offset" 
#define CS_PART_NAME	"a.cs%d_pblk_part%d_name" 

/* Eabling AUTOPIO operations */
#define USE_AUTOPIO

/* Prototype of routine that gets XENV */
extern int zxenv_get(char *recordname, void *dst, u32 *datasize);

/* Internal data structure */
static struct mtd_info smp86xx_mtds[MAX_CS];
static struct nand_chip smp86xx_chips[MAX_CS];
static struct mtd_partition *smp86xx_partitions[MAX_CS];
static struct nand_hw_control smp86xx_hw_control;
static int cs_avail[MAX_CS], cs_parts[MAX_CS];
static int cs_offset;

/* for 512B page, typically with 16B OOB */
static struct nand_ecclayout smp86xx_oobinfo_16 = {
	.eccbytes = 3,
	.eccpos = {10, 11, 12},
	.oobfree = {
		{.offset = 6, .length = 4},
		{.offset = 13, .length = 3},
	},
};

/* for 2KB page, typically with 64B OOB */
static struct nand_ecclayout smp86xx_oobinfo_64 = {
	.eccbytes = 12,
	.eccpos = {10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21},
	.oobfree = {
		{.offset = 6, .length = 4},
		{.offset = 22, .length = 42},
	},
};

/*
 * MTD structure for NAND controller
 */
const char *smp_nand_devname = SMP_NAND_DEV_NAME;

/**
 * smp86xx_read_byte -  read one byte from the chip
 * @mtd:	MTD device structure
 *
 *  read function for 8bit buswidth
 */
static u_char smp86xx_read_byte(struct mtd_info *mtd)
{
	struct nand_chip *this = mtd->priv;
	return RD_HOST_REG8((RMuint32)this->IO_ADDR_R + SMP86XX_REG_DATA);
}

#ifdef USE_AUTOPIO
static int mbus_done = 0;
static DECLARE_WAIT_QUEUE_HEAD(mbus_wq);
static void pbi_mbus_intr(int irq, void *arg)
{
	mbus_done = 1;
	wake_up_interruptible(&mbus_wq);
}
#endif

/**
 * smp86xx_write_buf -  write buffer to chip
 * @mtd:	MTD device structure
 * @buf:	data buffer
 * @len:	number of bytes to write
 *
 *  write function for 8bit buswidth
 */
static void smp86xx_write_buf(struct mtd_info *mtd, const u_char *buf, int len)
{
	int i;
	struct nand_chip *this = (struct nand_chip *)mtd->priv;

	//printk("%s smp86xx_write_buf buf = %p, len = 0x%08x(%d)\n", smp_nand_devname, buf, len, len);

#ifdef USE_AUTOPIO
	unsigned int cs = (u32)(this->IO_ADDR_R) / cs_offset;
	unsigned long g_mbus_reg = 0;

	if ((in_atomic()) || (len <= mtd->oobsize))
		goto pio;
	else if ((((u32)buf) < KSEG0) || (((u32)buf) >= KSEG2))
		goto pio;
	else if (em86xx_mbus_alloc_dma(SBOX_IDEFLASH, 0, &g_mbus_reg, NULL, 0))
		goto pio;

	dma_cache_wback_inv((unsigned long)buf, len);
	gbus_write_reg32(REG_BASE_host_interface + PB_automode_control + 4, 0);
	gbus_write_reg32(REG_BASE_host_interface + PB_automode_start_address, SMP86XX_REG_DATA); 
	gbus_write_reg32(REG_BASE_host_interface + PB_automode_control, (cs << 24) | (2 << 22) | (1 << 17) | (0 << 16) | len);

	em86xx_mbus_setup_dma(g_mbus_reg, tangox_dma_address(virt_to_phys(buf)), len, pbi_mbus_intr, NULL, 1);

	wait_event_interruptible(mbus_wq, mbus_done != 0);
	while (gbus_read_reg32(REG_BASE_host_interface + PB_automode_control) & 0xffff)
		; /* wait for AUTOPIO completion */
	mbus_done = 0;

	em86xx_mbus_free_dma(g_mbus_reg, SBOX_IDEFLASH);
	g_mbus_reg = 0;
	goto done;

pio:
#endif
	for (i = 0; i < len; i++) 
		WR_HOST_REG8((RMuint32)this->IO_ADDR_W + SMP86XX_REG_DATA, buf[i]);

#ifdef USE_AUTOPIO
done:
#endif
	return;
}

/**
 * smp86xx_read_buf -  read chip data into buffer
 * @mtd:	MTD device structure
 * @buf:	buffer to store date
 * @len:	number of bytes to read
 *
 *  read function for 8bit buswith
 */
static void smp86xx_read_buf(struct mtd_info *mtd, u_char *buf, int len)
{
	int i;
	struct nand_chip *this = mtd->priv;

#ifdef USE_AUTOPIO
	unsigned int cs = (u32)(this->IO_ADDR_R) / cs_offset;
	unsigned long g_mbus_reg = 0;

	if ((in_atomic()) || (len <= mtd->oobsize))
		goto pio;
	else if ((((u32)buf) < KSEG0) || (((u32)buf) >= KSEG2))
		goto pio;
	else if (em86xx_mbus_alloc_dma(SBOX_IDEFLASH, 1, &g_mbus_reg, NULL, 0))
		goto pio;

	dma_cache_inv((unsigned long)buf, len);
	gbus_write_reg32(REG_BASE_host_interface + PB_automode_control + 4, 0);
	gbus_write_reg32(REG_BASE_host_interface + PB_automode_start_address, SMP86XX_REG_DATA);
	gbus_write_reg32(REG_BASE_host_interface + PB_automode_control, (cs << 24) | (2 << 22) | (1 << 17) | (1 << 16) | len);

	em86xx_mbus_setup_dma(g_mbus_reg, tangox_dma_address(virt_to_phys(buf)), len, pbi_mbus_intr, NULL, 1);

	wait_event_interruptible(mbus_wq, mbus_done != 0);
	while (gbus_read_reg32(REG_BASE_host_interface + PB_automode_control) & 0xffff)
		; /* wait for AUTOPIO completion */
	mbus_done = 0;

	em86xx_mbus_free_dma(g_mbus_reg, SBOX_IDEFLASH);
	g_mbus_reg = 0;
	goto done;
pio:
#endif
	for (i = 0; i < len; i++) 
		buf[i] = RD_HOST_REG8((RMuint32)this->IO_ADDR_R + SMP86XX_REG_DATA);

#ifdef USE_AUTOPIO
done:
#endif
	return;
}

/**
 * smp86xx_verify_buf -  Verify chip data against buffer
 * @mtd:	MTD device structure
 * @buf:	buffer containing the data to compare
 * @len:	number of bytes to compare
 *
 *  verify function for 8bit buswith
 */
static int smp86xx_verify_buf(struct mtd_info *mtd, const u_char *buf, int len)
{
#ifdef USE_AUTOPIO
	u_char tmpbuf[2048];	/* up to 2KB */
	smp86xx_read_buf(mtd, tmpbuf, len);
	return (memcmp(buf, tmpbuf, len) == 0) ? 0 : -EIO;
#else
	int i;
	struct nand_chip *this = mtd->priv;
	for (i = 0; i < len; i++) {
		if (buf[i] != RD_HOST_REG8((RMuint32)this->IO_ADDR_R + SMP86XX_REG_DATA))
			return -EIO;
	}
	return 0;
#endif
}

/* smp86xx_nand_hwcontrol
 *
 * Issue command and address cycles to the chip
 */
static void smp86xx_nand_hwcontrol(struct mtd_info *mtd, int cmd, unsigned int ctrl)
{
	register struct nand_chip *this = mtd->priv;

	if (cmd == NAND_CMD_NONE)
		return;

	if (ctrl & NAND_CLE)
		WR_HOST_REG8((RMuint32)this->IO_ADDR_W + SMP86XX_REG_CMD, cmd);
	else
		WR_HOST_REG8((RMuint32)this->IO_ADDR_W + SMP86XX_REG_ADDR, cmd);
}

/* smp86xx_nand_devready()
 *
 * returns 0 if the nand is busy, 1 if it is ready
 */
static int smp86xx_nand_devready(struct mtd_info *mtd)
{
	return RD_HOST_REG32(PB_CS_ctrl) & PB_IORDY;
}

/* ECC handling functions */

static inline int mu_count_bits(u32 v)
{
	int i, count;
	for (count = i = 0; (i < 32) && (v != 0); i++, v >>= 1)
		count += (v & 1);
	return count;
}

/* correct 512B packet */
static int ecc_correct_512(struct mtd_info *mtd, u_char *dat,
				     u_char *read_ecc, u_char *calc_ecc)
{
	u32 mem, reg;
	mem = read_ecc[0] | ((read_ecc[1] & 0x0f) << 8) | ((read_ecc[1] & 0xf0) << 12) | (read_ecc[2] << 20);
	reg = calc_ecc[0] | ((calc_ecc[1] & 0x0f) << 8) | ((calc_ecc[1] & 0xf0) << 12) | (calc_ecc[2] << 20);
	if (mem == reg)
		return 0;
	else {
		u16 pe, po, is_ecc_ff;
		is_ecc_ff = ((mem & 0x0fff0fff) == 0x0fff0fff);
		mem ^= reg;

		switch(mu_count_bits(mem)) {
			case 0:
				return 0;
			case 1:
				return -1;
			case 12:
				po = (u16)(mem & 0xffff);
				pe = (u16)((mem >> 16) & 0xffff);
				po = pe ^ po;
				if (po == 0x0fff) {
					dat[pe >> 3] ^= (1 << (pe & 7));
					return 1;	/* corrected data */
				} else 
					return -1;	/* failed to correct */
			default:
				return (is_ecc_ff && (reg == 0)) ? 0 : -1;
		}
	}
	return -1;	/* should not be here */
}

/* correct 512B * 4 packets */
static int ecc_correct_2048(struct mtd_info *mtd, u_char *dat,
				     u_char *read_ecc, u_char *calc_ecc)
{
	int ret0, ret1, ret2, ret3;
	
	ret0 = ecc_correct_512(mtd, dat, read_ecc, calc_ecc);
	ret1 = ecc_correct_512(mtd, dat + 512, read_ecc + 3, calc_ecc + 3);
	ret2 = ecc_correct_512(mtd, dat + 1024, read_ecc + 6, calc_ecc + 6);
	ret3 = ecc_correct_512(mtd, dat + 1536, read_ecc + 9, calc_ecc + 9);

	if ((ret0 < 0) || (ret1 < 0) || (ret2 << 0) || (ret3 << 0))
		return -1;
	else
		return ret0 + ret1 + ret2 + ret3;
}

/* ECC functions
 *
 * These allow the smp86xx to use the controller's ECC
 * generator block to ECC the data as it passes through]
*/
static int smp86xx_nand_calculate_ecc_512(struct mtd_info *mtd, const u_char *dat, u_char *ecc_code)
{
	ecc_code[0] = RD_HOST_REG8(PB_ECC_code0 + 0);
	ecc_code[1] = RD_HOST_REG8(PB_ECC_code0 + 1);
	ecc_code[2] = RD_HOST_REG8(PB_ECC_code0 + 2);

	pr_debug("%s: returning ecc %02x%02x%02x\n", __func__,
		 ecc_code[0], ecc_code[1], ecc_code[2]);

	return 0;
}

static int smp86xx_nand_calculate_ecc_2048(struct mtd_info *mtd, const u_char *dat, u_char *ecc_code)
{
	ecc_code[0] = RD_HOST_REG8(PB_ECC_code0 + 0);
	ecc_code[1] = RD_HOST_REG8(PB_ECC_code0 + 1);
	ecc_code[2] = RD_HOST_REG8(PB_ECC_code0 + 2);
	pr_debug("%s: returning ecc %02x%02x%02x", __func__, ecc_code[0], ecc_code[1], ecc_code[2]);

	ecc_code[3] = RD_HOST_REG8(PB_ECC_code1 + 0);
	ecc_code[4] = RD_HOST_REG8(PB_ECC_code1 + 1);
	ecc_code[5] = RD_HOST_REG8(PB_ECC_code1 + 2);
	pr_debug("%02x%02x%02x", ecc_code[3], ecc_code[4], ecc_code[5]);

	ecc_code[6] = RD_HOST_REG8(PB_ECC_code2 + 0);
	ecc_code[7] = RD_HOST_REG8(PB_ECC_code2 + 1);
	ecc_code[8] = RD_HOST_REG8(PB_ECC_code2 + 2);
	pr_debug("%02x%02x%02x", ecc_code[6], ecc_code[7], ecc_code[8]);

	ecc_code[9] = RD_HOST_REG8(PB_ECC_code3 + 0);
	ecc_code[10] = RD_HOST_REG8(PB_ECC_code3 + 1);
	ecc_code[11] = RD_HOST_REG8(PB_ECC_code3 + 2);
	pr_debug("%02x%02x%02x\n", ecc_code[9], ecc_code[10], ecc_code[11]);

	return 0;
}

static void smp86xx_nand_enable_hwecc(struct mtd_info *mtd, int mode)
{
	WR_HOST_REG32(PB_ECC_clear, 0);
}

/* Loading partiton information from XENV */
static void smp86xx_nand_load_part_info(void)
{
#define BUFSIZE		256
	char buf[BUFSIZE], pname[BUFSIZE];
	u32 poff, psz, dsize, rsvd_blks;
	u32 cs, part, parts, cnt, blkmask;
	u64 rsvd_sz;

	for (cs = 0; cs < MAX_CS; cs++) {

		if (cs_avail[cs] == 0)	/* not available */
			continue;

		sprintf(buf, CS_RESERVED, cs);
		dsize = sizeof(u32);
		if ((zxenv_get(buf, &rsvd_blks, &dsize) < 0) || (dsize != sizeof(u32))) {
			cs_avail[cs] = 0;
			continue;
		} 

		/* find out the size of reservation zone */
		smp86xx_mtds[cs].size = rsvd_sz = (u64)smp86xx_mtds[cs].erasesize * (u64)rsvd_blks;
		blkmask = ~(smp86xx_mtds[cs].erasesize - 1);

		sprintf(buf, CS_PARTS, cs);
		dsize = sizeof(u32);
		if ((zxenv_get(buf, &parts, &dsize) < 0) || (dsize != sizeof(u32)))
			continue;
		else if (parts > MAX_PARTITIONS)
			parts = MAX_PARTITIONS;

		if ((smp86xx_partitions[cs] = kmalloc(sizeof(struct mtd_partition) * parts, GFP_KERNEL)) == NULL) {
			cs_avail[cs] = 0;
			continue;
		}
		memset(smp86xx_partitions[cs], 0, sizeof(struct mtd_partition) * parts);

		for (part = cnt = 0; part < parts; part++) {
			
			sprintf(buf, CS_PART_SIZE, cs, part + 1);
			dsize = sizeof(u32);
			if ((zxenv_get(buf, &psz, &dsize) < 0) || (dsize != sizeof(u32)))
				goto next;
			psz &= blkmask;		/* make it align to block boundary */
			if (psz == 0)
				goto next;
			smp86xx_partitions[cs][cnt].size = psz;

			sprintf(buf, CS_PART_OFFSET, cs, part + 1);
			dsize = sizeof(u32);
			if ((zxenv_get(buf, &poff, &dsize) < 0) || (dsize != sizeof(u32)))
				goto next;
			else if ((poff & (smp86xx_mtds[cs].erasesize - 1)) != 0)	/* not aligned to block boundary */
				goto next;
			smp86xx_partitions[cs][cnt].offset = poff;

			/* check if partition is out of reservation zone */
			if (((u64)poff >= rsvd_sz) || (((u64)poff + (u64)psz) > rsvd_sz))
				goto next;

			sprintf(buf, CS_PART_NAME, cs, part + 1);
			dsize = BUFSIZE;
			memset(pname, 0, BUFSIZE);
			if ((zxenv_get(buf, pname, &dsize) == 0) && (dsize > 0)) {	/* partition name is given */
				char *p;
				u32 i;
				for (i = 0, p = pname; (*p != '\0') && (i < dsize); p++, i++) {
					if (!isspace(*p))
						break;
				}
				if (*p == '\"') {	/* found leading '\"', try strip out trailing '\"' */
					char *e;
					p++;
					for (i = strnlen(p, BUFSIZE - i), e = p + (i - 1); i > 0; i--, e--) {
						if (isspace(*e))
							*e = '\0';
						else {
							if (*e == '\"') 
								*e = '\0';
							break;
						}
					}
				}
				smp86xx_partitions[cs][cnt].name = kmalloc(strnlen(p, BUFSIZE) + 1, GFP_KERNEL);
				if (smp86xx_partitions[cs][cnt].name)
					strncpy(smp86xx_partitions[cs][cnt].name, p, strnlen(p, BUFSIZE) + 1);
				else
					goto next;
			} else {	/* cooked-up partition name here */
				sprintf(buf, "CS%d/PBPart%d", cs, part + 1);
				smp86xx_partitions[cs][cnt].name = kmalloc(16, GFP_KERNEL);
				if (smp86xx_partitions[cs][cnt].name)
					strncpy(smp86xx_partitions[cs][cnt].name, buf, 16);
				else
					goto next;
			}

			cnt++;
			continue;	/* next partition */
next:
			smp86xx_partitions[cs][cnt].size = smp86xx_partitions[cs][cnt].offset = 0;
		}

		cs_parts[cs] = cnt;
	}
}

static int __init smp86xx_nand_init(void)
{
	struct nand_chip *this;
	u32 mem_staddr;
	RMuint8 local_pb_cs_ctrl;
	RMuint32 local_pb_cs_config, local_pb_cs_config1;
	int cs, chip_cnt = 0;

	memset(smp86xx_mtds, 0, sizeof(struct mtd_info) * MAX_CS);
	memset(smp86xx_chips, 0, sizeof(struct nand_chip) * MAX_CS);
	memset(smp86xx_partitions, 0, sizeof(struct mtd_partition *) * MAX_CS);
	memset(cs_avail, 0, sizeof(int) * MAX_CS);
	memset(cs_parts, 0, sizeof(int) * MAX_CS);

	printk("%s SMP86xx NAND Driver %s\n", smp_nand_devname, SMP_NAND_DRV_VERSION);
	
	local_pb_cs_ctrl = RD_HOST_REG8(PB_CS_ctrl);
	local_pb_cs_config = RD_HOST_REG32(PB_CS_config);
	local_pb_cs_config1 = RD_HOST_REG32(PB_CS_config1);
	switch((local_pb_cs_ctrl >> 4) & 7) {
		case 0: cs_offset = 0x200;
			break;
		case 1: 
		case 2: cs_offset = 0x100;
			break;
		default:
			printk("No NAND flash is available (0x%x).\n", local_pb_cs_ctrl);
			return -EIO;
	}

	for (cs = 0; cs < 4; cs++) {
		if ((local_pb_cs_config >> (20 + cs)) & 1)
			cs_avail[cs] = 1;
	}
#if (MAX_CS >= 4)
	for (cs = 0; cs < 4; cs++) {
		if ((local_pb_cs_config1 >> (20 + cs)) & 1)
			cs_avail[cs + 4] = 1;
	}
#endif

	spin_lock_init(&smp86xx_hw_control.lock);
	init_waitqueue_head(&smp86xx_hw_control.wq);

	for (cs = 0; cs < MAX_CS; cs++) {
		if (cs_avail[cs] == 0)
			goto next;

		smp86xx_mtds[cs].priv = &smp86xx_chips[cs];
		smp86xx_mtds[cs].owner = THIS_MODULE;
		this = &smp86xx_chips[cs];

		mem_staddr = cs * cs_offset;

		/* 30 us command delay time */
		this->chip_delay = 30;
		this->ecc.mode = NAND_ECC_SOFT;
		this->options = NAND_NO_AUTOINCR | BBT_AUTO_REFRESH;
		this->controller = &smp86xx_hw_control;

		this->read_byte = smp86xx_read_byte;
		this->read_buf = smp86xx_read_buf;
		this->write_buf = smp86xx_write_buf;
		this->verify_buf = smp86xx_verify_buf;

		this->cmd_ctrl     = smp86xx_nand_hwcontrol;
		this->dev_ready    = smp86xx_nand_devready;

		this->IO_ADDR_W	= this->IO_ADDR_R = (void __iomem *)mem_staddr;

		/* nand reset */
		WR_HOST_REG8((RMuint32)this->IO_ADDR_W + SMP86XX_REG_CMD, NAND_CMD_RESET);
		udelay(tReset);

		printk("%s: checking NAND device on CS%d ..\n", SMP_NAND_DEV_NAME, cs);

		/* Scan to find existence of the device */
		if (nand_scan_ident(&smp86xx_mtds[cs], 1)) 
			goto next;

		this->ecc.mode = NAND_ECC_HW;
		this->ecc.steps = 1;
		this->ecc.hwctl = smp86xx_nand_enable_hwecc;
		switch (smp86xx_mtds[cs].writesize) {
			case 512:
				this->ecc.calculate = smp86xx_nand_calculate_ecc_512;
				this->ecc.correct = ecc_correct_512;
				this->ecc.size = 512;
				this->ecc.bytes = this->ecc.total = 3;
				this->ecc.layout = &smp86xx_oobinfo_16;
				break;
			case 2048:
				this->ecc.calculate = smp86xx_nand_calculate_ecc_2048;
				this->ecc.correct = ecc_correct_2048;
				this->ecc.size = 2048;
				this->ecc.bytes = this->ecc.total = 12;
				this->ecc.layout = &smp86xx_oobinfo_64;
				break;
			default:
				printk("%s: unsupported NAND on CS%d.\n", SMP_NAND_DEV_NAME, cs);
				goto next;
		}

		if (nand_scan_tail(&smp86xx_mtds[cs]))
			goto next;

		printk("%s: detected NAND on CS%d, erasesize %d, pagesize %d, oobsize %d, oobavail %d\n", 
			SMP_NAND_DEV_NAME, cs, smp86xx_mtds[cs].erasesize,
			smp86xx_mtds[cs].writesize, smp86xx_mtds[cs].oobsize, smp86xx_mtds[cs].oobavail);
		cs_avail[cs] = 1;
		chip_cnt++;
		continue;
next:
		cs_avail[cs] = 0;
		continue;
	}

	if (chip_cnt) {
		printk("%s: detection completed, load partition information from XENV ..\n", SMP_NAND_DEV_NAME);
		smp86xx_nand_load_part_info();
		for (cs = 0; cs < MAX_CS; cs++) {
			/* Register the partitions */
			if (cs_avail[cs]) {
				add_mtd_device(&smp86xx_mtds[cs]);
				if (cs_parts[cs]) {
					printk("%s: load partition information for CS%d ..\n", SMP_NAND_DEV_NAME, cs);
					add_mtd_partitions(&smp86xx_mtds[cs], smp86xx_partitions[cs], cs_parts[cs]);
				}
			}
		}
		return 0;
	} else {
		printk("No NAND flash is detected.\n");
		return -EIO;
	}
}

static void __exit smp86xx_nand_exit(void)
{
	int cs, i;

	for (cs = 0; cs < MAX_CS; cs++) {
		/* Release resources, unregister device */
		if (cs_avail[cs]) {
			nand_release(&smp86xx_mtds[cs]);
			if (smp86xx_partitions[cs]) {
				for (i = 0; i < cs_parts[cs]; i++) {
					if (smp86xx_partitions[cs][i].name)
						kfree(smp86xx_partitions[cs][i].name);
				}
				kfree(smp86xx_partitions[cs]);
			}
		}
	}
}

module_init(smp86xx_nand_init);
module_exit(smp86xx_nand_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Ilhong Yoon <iyoon@sdesigns.com>");
MODULE_DESCRIPTION("SMP86xx MTD NAND driver");

