#!/bin/bash
trap "cleanup" EXIT SIGTERM SIGINT

if [ -z "$XLOAD3_ROOT" ]; then
	export XLOAD3_ROOT=/utils/em8xxx/signed_items/xload3
fi
if [ -z "$XLOAD3_GEN_SCRIPTS" ]; then
	export XLOAD3_GEN_SCRIPTS=$XLOAD3_ROOT/scripts/gen
fi

pushd `dirname $0` &> /dev/null
scripts=`pwd -P`
popd &> /dev/null
export PATH=$PATH:$scripts
source $scripts/xload3.env

# Signing config file parameters
CFG_CKDS=''
CFG_CERT_PRIVATE_KEYS=''
CFG_CERT=''

# Encryption config file parameters
ECFG_CIPHER=''
ECFG_MODE=''
ECFG_IDX0=''
ECFG_IDX1=''
ECFG_IDX2=''
ECFG_KEY=''
ECFG_SEED=''

# globals
tmpdir=''
cfg=''
ecfg=''
payload=''
p=''
certid=''
c=''
xload3out=''
decrypted_private_keys=()
ndecrypted=0

USAGE=`cat<<EOF
Usage: $0 --payload <payload> (--certid <certid> | --cfg <cfgfile>)
          [--ecfg <ecfgfile>]

  payload: (required)
     Path to payload file.

  certid:
     16-bit certificate id.  Specify to use a default config
     file corresponding to this certificate id.

  cfgfile:
     Signing Configuration file that specifies certificate keys
     to use.

  ecfgfile:
     Encryption Configuration file that specifies parameters for
     encrypting the payload.  If left unspecified the payload will
     be left unencrypted.
EOF`

# Override logfile writing (locally only)
function log_interface_logfile() {
	echo -n "" >&2
}

function usage() {
	if [ ! -z "$1" ]; then logerr "$@"; fi
	echo -e "$USAGE\n" >&2 
	exit -1;
}

function cleanup() {
	for f in ${decrypted_private_keys[@]}; do
		if [ -f $f ]; then
			echo "Removing decrypted private key $f"
			shred -u $f
		fi
	done
	if [ -d "$tmpdir" ]; then
		echo ""
		rm -rf $tmpdir
	fi
}

function parse_cmdline() {

    while [ "$1" ]; do

        # found & params_consumed are globals set by
        # check_and_set_* methods
        found=0
        params_consumed=0
		set_param_marker '--'
		check_and_set_file_param "payload" $@
		check_and_set_int_param "certid" 0 0xffff $@
		check_and_set_file_param "cfg" $@
		check_and_set_file_param "ecfg" $@

		# Maintain backwards compatability with old api '-p' and '-c'
		set_param_marker '-'
		check_and_set_file_param "p" $@
		if [ ! -z "$p" ]; then payload=$p; p=''; fi
		check_and_set_int_param "c" 0 0xffff $@
		if [ ! -z "$c" ]; then certid=$c; c=''; fi

        if [ $found -eq 0 ]; then
            usage "Error: invalid parameter!"
        fi

        # shift out used parameters
        for i in `seq 1 $params_consumed`; do shift; done
    done

	# check that all required params have been set:
	exit_if_not_defined "payload"
	if [ "$certid" ]; then
		# certid should be 4-byte hex formatted string
		certid=`printf "%04x" $certid`
	else
		msg="Either 'certid' or 'cfg' must be specified."
		warn_if_not_defined "certid"
		if [ -z "$certid" ]; then
			exit_if_not_defined "cfg" "$msg"
		else
			warn_if_not_defined "cfg"
		fi
	fi
	warn_if_not_defined "ecfg" "Payload will be left unencrypted."
}

function validate_cfg_parameters() {
	if [ ! -f "$CFG_CERT" ]; then
		logerr "undefined cert in cfg file ($cfg)"
		exit -1;
	fi
	if [ -z "$CFG_CKDS" ]; then
		logerr "no ckds defined within cfg file ($cfg)"
		exit -1;
	fi
	if [ -z "$CFG_CERT_PRIVATE_KEYS" ]; then
		logerr "no private keys defined within cfg file ($cfg)"
		exit -1;
	fi
	for ckd in ${CFG_CKDS[@]}; do
		if [ ! -f "$ckd" ]; then
			logerr "invalid ckd found ($ckd)!"
			exit -1;
		fi
		echo "CKD--: $ckd"
	done
	for k in ${CFG_CERT_PRIVATE_KEYS[@]}; do
		if [ ! -f "$k" ]; then
			logerr "invalid private key found ($k)!"
			exit -1;
		fi
		echo "key--: $k"
	done
	if [ ${#CFG_CKDS[@]} -ne ${#CFG_CERT_PRIVATE_KEYS[@]} ]; then
		logerr "mismatch between number of Cert keys and CKDs"
		exit -1;
	fi
}

function validate_ecfg_parameters() {
	case $ECFG_CIPHER in
	none)
		echo "No cipher used."
		return 0;
		;;
	aes128)
		echo "Using aes-128-cbc for cipher."
		;;
	*)
		logerr "invalid cipher ($ECFG_CIPHER)."
		return 1;
		;;
	esac

	case $ECFG_MODE in 
	1)
		log "Using direct encryption (mode1)."
		if [ -z $ECFG_IDX0 ]; then
			logerr "missing idx0 for mode1 encryption!"
			return 1;
		fi
		;;
	2)
		log "Using indirect encryption (mode2)."
		if [[ -z $ECFG_IDX0 || -z $ECFG_IDX1 ]]; then
			logerr "missing idx0 or idx1 for mode2 encryption!"
			return 1;
		echo ""
		fi
		;;
	3)
		log "Using double-indirect encryption (mode3)."
		if [[ -z $ECFG_IDX0 || -z $ECFG_IDX1 || -z $ECFG_IDX2 ]]; then
			logerr "missing idx0 or idx1 or idx2 for mode2 encryption!"
			return 1;
		fi
		;;
	*)
		logerr "Invalid mode specified ($ECFG_MODE)"
		return 1;
		;;
	esac

	if [[ -z "$ECFG_KEY" || ! -f "$ECFG_KEY" ]]; then
		logerr "Missing encryption key ($ECFG_KEY)!"
		return 1;
	fi
}

## Start of script execution

source $scripts/process_args.env
parse_cmdline $@
cert_dir=$XLOAD3_SIGNED_CERTS

if [ -z "$cfg" ]; then
	# Hack for testing:
	cfg=`ls $cert_dir/config/*$certid* 2> /dev/null`;
fi

if [ ! -f "$cfg" ]; then
	logerr "Signing configuration file not found! ($cfg)"
	exit -1;
else
	echo "Using configuration file: $cfg"
fi
source $cfg
if ! validate_cfg_parameters; then exit -1; fi
cert=$CFG_CERT

tmpdir=`mktemp -d tmp.xload3.sign.XXXXXX`
pl=$tmpdir/`basename $payload`
cp $payload $pl
	
# strip '.bin' from the end of the file, replace with '.xload3'
xload3filename=`basename $payload|sed 's/\(.*\)\.bin$/\1/'`.xload3
# Default to same directory with the input payload.
xload3out=`dirname $payload`/$xload3filename

# Add encryption parameters if ecfg specified
if [ "$ecfg" ]; then
	source $ecfg
	if ! validate_ecfg_parameters; then exit -1; fi
	payload_params="--cipher $ECFG_CIPHER"

	if [ "$ECFG_MODE" ]; then
		payload_params="$payload_params --mode $ECFG_MODE"
	fi
	if [ "$ECFG_IDX0" ]; then
		payload_params="$payload_params --idx0 $ECFG_IDX0"
	fi
	if [ "$ECFG_IDX1" ]; then
		payload_params="$payload_params --idx1 $ECFG_IDX1"
	fi
	if [ "$ECFG_IDX2" ]; then
		payload_params="$payload_params --idx2 $ECFG_IDX2"
	fi
	if [ "$ECFG_SIZE" ]; then
		payload_params="$payload_params --size $ECFG_SIZE"
	fi
	
	if [ "$ECFG_OUT_DIR" ]; then
		xload3out=$ECFG_OUT_DIR/$xload3filename
	fi

	key=$ECFG_KEY
	if [ ${key:${#key}-4:4} = ".gpg" ]; then
		echo "Encryption key ($key) is gpg encrypted."
		keyname=`basename ${key:0:${#key}-4}`
		key=`mktemp $tmpdir/tmp.DECRYPTED_ENCRYPTION_KEY.$keyname.XXXXXX`
		gpg -o $key --yes ${ECFG_KEY}
		decrypted_private_keys[$ndecrypted]=$key
		((ndecrypted++))
	elif [ ${key:${#key}-6:6} = ".smime" ]; then
		# Assume it is protected by ncipher.
		echo "Signing key ($key) is smime encrypted."
		protectedkey=$tmpdir/`basename $ECFG_KEY`
		cp $ECFG_KEY $protectedkey
		key=${protectedkey:0:${#protectedkey}-6}
		$XLOAD3_NCIPHER_SCRIPTS/decrypt_protected_file.bash $key
		decrypted_private_keys[$ndecrypted]=$key
		((ndecrypted++))
	fi
	payload_params="$payload_params --key $key"
	
	if [ "$ECFG_SEED" ]; then
		payload_params="$payload_params --seed $ECFG_SEED"
	fi
else
	payload_params="--cipher none"
fi

if [ ! -w `dirname $xload3out` ]; then
	logerr "No write permission to create file $xload3out."
	exit -1;
fi

echo "cat $pl | $XLOAD3_GEN_SCRIPTS/create_payload.pl $payload_params \
		--hdr $pl.hdr --out $pl.enc"
cat $pl | $XLOAD3_GEN_SCRIPTS/create_payload.pl $payload_params \
		--hdr $pl.hdr --out $pl.enc

# sign with all cert keys
for i in `seq 0 $((${#CFG_CKDS[@]} - 1))`; do
	key=${CFG_CERT_PRIVATE_KEYS[$i]}
	ckd=${CFG_CKDS[$i]}

	if [ ${key:${#key}-4:4} = ".gpg" ]; then
		echo "Signing key ($key) is gpg encrypted."
		keyname=`basename ${key:0:${#key}-4}`
		key=`mktemp $tmpdir/tmp.DECRYPTED_CERT_PRIVATE_KEY.$keyname.XXXXXX`
		gpg -o $key --yes ${CFG_CERT_PRIVATE_KEYS[$i]}
		decrypted_private_keys[$ndecrypted]=$key
		((ndecrypted++))
	fi
	
	echo "Signing with key=$key ckd=$ckd"

	# sign the header
	$XLOAD3_GEN_SCRIPTS/create_payload_sig.pl --pem $key \
			--cert-key ${CFG_CKDS[$i]} \
			--payload $pl.hdr > $pl.hdr.sig.$i
	# sign the payload
	$XLOAD3_GEN_SCRIPTS/create_payload_sig.pl --pem $key \
			--cert-key $ckd \
			--payload $pl > $pl.sig.$i
done

# output hdr + all hdr sigs + encrypted payload + all payload sigs
# note: if there are more than 10 keys, then there will be a problem
#       with the ordering of the signatures (shouldn't happen!)
cat $cert $pl.hdr $pl.hdr.sig* $pl.enc $pl.sig* > $xload3out

# For now maintain backwards compatability (.bin.xload3).
# TODO: REMOVEME
cp $xload3out `dirname $xload3out`/`basename $payload`.xload3

rm -rf $tmpdir

